# Kornpob Bhirombhakdi 20200331

import numpy as np

def make_SIP(coef,x,y,startx=True):
    """
    Simple imaging polynomial (SIP) is a conventional method to describe non-linear variation in an image. Ref: https://fits.gsfc.nasa.gov/registry/sip/shupeADASS.pdf.
    ##########
    Assume a SIP model of order 2, i.e., Z = a0 + a1*X + a2*X**2.
    Typically, X is relative to SIP reference system whose origin is corresponding to (xref,yref) in the original image. Therefore, X = x - xref where (x,y) is an image pixel.
    Z is a quantity of interests. In aXe grism reduction, Z can be Y (as y = Y + yref) for trace, or wavelength.
    SIP coefficients are 2D with a given polynomial order. Assume the order is 3. Therefore, ai = ai0 + ai1*X' + ai2*Y' + ai3*X'**2 + ai4*X'*Y' + ai5*Y'**2 + ... + ai9*Y'**3.
    Note X is the leading term (this is specified by startx=True in make_SIP). Set startx=False otherwise.
    Note that X and X' might be different. For aXe reduction, (X',Y') = (xd,yd) as the source location from direct image.
    """
    if startx:
        xref,yref = x,y
    else:
        xref,yref = y,x
    n = len(coef)
    d = []
    px,py = 0,0
    a = [(px,py)]
    b = [(xref,yref)]
    p = 0
    q = True
    while(q):
        if px==0:
            p+=1
            px=p
            py=0
        else:
            px-=1
            py+=1
        a.append((px,py))
        b.append((xref,yref))
        if len(a)>=len(coef):
            q = False
    a,b = np.array(a),np.array(b)
    c = b**a
    c = np.sum(c[:,0]*c[:,1]*coef)
    d.append(c)
    d = np.array(d)
    return d 



# Kornpob Bhirombhakdi

import numpy as np
import matplotlib.pyplot as plt
import copy,glob,os
from astropy.io import fits
from polynomial2d.polynomial2d import Polynomial2D
from math import pi

class AXEhelper_POLY2D:
    def __init__(self,axehelperbkg):
        self.axehelperbkg = axehelperbkg
        self.obj = self._obj()
    ####################
    ####################
    ####################
    def saveflt(self,savepath='./EXTRA/'):
        self._check(savepath)
        fltflist = self.axehelperbkg.fltflist
        axeflist = self.axehelperbkg.axeflist
        OBJ = self.obj
        HEADERALL = self.axehelperbkg.headerall
        for ii,i in enumerate(fltflist):
            root = i.split('/')[-1].split('_')[0]
            string = '{0}_flt.fits'.format(root)
            string = savepath+string
            string2 = '{0}_bkg.fits'.format(root)
            string2 = savepath+string2
    
            os.system('cp {0} {1}'.format(i,string))
            os.system('cp {0} {1}'.format(i,string2))

            obj = OBJ[i]

            tmpheader = HEADERALL[axeflist[ii]]
            xref,yref = tmpheader['XYREF']
            pixx = tmpheader['PIXX']
            pixy = tmpheader['PIXY']
            ww = tmpheader['WW']
            cc0x,cc0y,cc1x,cc1y = tmpheader['CC']
            sectx = tmpheader['SECTX']
            secty = tmpheader['SECTY']

            tmp = fits.open(string)
            tmp[1].data[cc0y:cc1y,cc0x:cc1x] = obj.model['YSUB'].copy()
            tmp.writeto(string,overwrite=True)
            tmp.close()

            tmp = fits.open(string2)
            tmp[1].data[cc0y:cc1y,cc0x:cc1x] = obj.model['YFIT'].copy()
            tmp.writeto(string2,overwrite=True)
            tmp.close()
    def _check(self,savepath):
        try:
            os.mkdir(savepath)
        except:
            try:
                cwd = os.getcwd()
                os.chdir(savepath)
                os.chdir(cwd)
            except:
                raise ValueError('cannot mkdir {0}'.format(savepath))        
    ####################
    ####################
    ####################
    def fit(self,norder=4):
        OBJ = self.obj
        for ii,i in enumerate(OBJ):
            obj = OBJ[i]
            obj.model['NORDER'] = norder
            obj.fit()
            obj.compute()

            obj.model['YSUB'] = obj.data['Y'] - obj.model['YFIT']
            obj.model['YSUB_MASK'] = obj.model['YSUB'] * obj.data['MASKOBJ'].astype(int)
            obj.model['SUM'] = obj.model['YSUB_MASK'].sum(axis=0)

            OBJ[i] = copy.deepcopy(obj)   
        self.obj = OBJ
    ####################
    ####################
    ####################  
    def _obj(self):
        OBJ = {}
        axehelperbkg = self.axehelperbkg
        fltflist = axehelperbkg.fltflist
        axeflist = axehelperbkg.axeflist
        HEADERALL = axehelperbkg.headerall
        for ii,i in enumerate(fltflist):
            # image
            tmp = fits.open(i)
            tmpdata = tmp[1].data.copy()
            tmpdq = tmp['DQ'].data.copy()

            # header & prep
            tmpheader = HEADERALL[axeflist[ii]]
            xref,yref = tmpheader['XYREF']
            pixx = tmpheader['PIXX']
            pixy = tmpheader['PIXY']
            cc0x,cc0y,cc1x,cc1y = tmpheader['CC']
            sectx = tmpheader['SECTX']
            secty = tmpheader['SECTY']

            # Polynomial2D
            x1 = np.arange(cc0x,cc1x)
            x2 = np.arange(cc0y,cc1y)
            x1,x2 = np.meshgrid(x1,x2)

            obj = Polynomial2D()
            obj.data['X1'] = x1.copy()
            obj.data['X2'] = x2.copy()
            obj.data['Y'] = tmpdata[cc0y:cc1y,cc0x:cc1x]
        #     print(x1.shape,x2.shape,obj.data['Y'].shape)    

            # data['MASK']
            tmp = np.full_like(tmpdq,True,dtype=bool)
            m = np.where(tmpdq==0)
            tmp[m] = False
            obj.data['MASK'] = tmp[cc0y:cc1y,cc0x:cc1x]
        #     print(obj.data['Y'].shape,obj.data['MASK'].shape) 
        
            # MASKOBJ
            tmp = np.full_like(obj.data['MASK'],False,dtype=bool)
            bb0x = sectx[0]+sectx[1]
            bb1x = bb0x+sectx[2]
            bb0y = secty[0]
            bb1y = bb0y+secty[1]+secty[2]+secty[3]   
            tmp[bb0y:bb1y,bb0x:bb1x] = True
            obj.data['MASKOBJ'] = tmp.copy()
            
            # MASKFIT
            obj.model['MASKFIT'] = obj.data['MASKOBJ'] | obj.data['MASK']

            # output
            OBJ[i] = copy.deepcopy(obj)
        return OBJ     
    ####################
    ####################
    ####################
    def show2d(self,save=False,savefname='default'):
        OBJ = self.obj
        axeflist = self.axehelperbkg.axeflist
        HEADERALL = self.axehelperbkg.headerall
        for ii,i in enumerate(OBJ):
            obj = OBJ[i]
            tmpdata = obj.data['Y']
            tmpmask = obj.data['MASK']
            tmpheader = HEADERALL[axeflist[ii]]
            xref,yref = tmpheader['XYREF']
            pixx = tmpheader['PIXX']
            pixy = tmpheader['PIXY']
            ww = tmpheader['WW']
            cc0x,cc0y,cc1x,cc1y = tmpheader['CC']
            sectx = tmpheader['SECTX']
            secty = tmpheader['SECTY']

            # shift due to the cut
            tmpxref = xref - cc0x
            tmpyref = yref - cc0y
            tmppixx = pixx - cc0x
            tmppixy = pixy - cc0y

            fig,ax = plt.subplots(2,1,sharex=True)
            m = np.where(np.isfinite(tmpdata))
            vmin,vmax = np.percentile(tmpdata[m],5.),np.percentile(tmpdata[m],99.)
            ax[0].imshow(tmpdata,origin='lower',cmap='viridis',vmin=vmin,vmax=vmax)
            ax[0].scatter(tmpxref,tmpyref,s=30,facecolor='red',edgecolor='None')
            ax[0].plot(tmppixx,tmppixy,'r-')
            ax[0].set_title('{0}'.format(i.split('/')[-1].split('_')[0]),fontsize=20)
            ax[0].set_ylabel('pixY + {0}'.format(cc0y),fontsize=20)

            bb0x = sectx[0]+sectx[1]
            bb1x = bb0x+sectx[2]
            bb0y = secty[0]
            bb1y = bb0y+secty[1]+secty[2]+secty[3]
            tmpx = [bb0x,bb1x,bb1x,bb0x,bb0x]
            tmpy = [bb0y,bb0y,bb1y,bb1y,bb0y]
            ax[0].plot(tmpx,tmpy,'r-')

            ax[1].plot(tmppixx,ww)
            ax[1].set_xlabel('pixX + {0}'.format(cc0x),fontsize=20)
            ax[1].set_ylabel('obs. wavelength (A)',fontsize=20)
            ax[1].grid()
            
            if save:
                if savefname=='default':
                    string = '/'.join(axeflist[ii].split('/')[0:-1])
                    string += '/{0}_axehelperpoly2dshow2d.png'.format(axeflist[ii].split('/')[-1].split('.')[0])
                else:
                    string = savefname
                fig.savefig(string,bbox_inches='tight')
    ####################
    ####################
    ####################
    def show3d(self,save=False,savefname='default'):
        OBJ = self.obj
        axeflist = self.axehelperbkg.axeflist
        for ii,i in enumerate(OBJ):
            obj = OBJ[i]
            string = i

            fig = plt.figure(figsize=(10,10))
            ax = plt.axes(projection='3d')
            tmp = copy.deepcopy(obj.data)
            ax.plot_surface(tmp['X1'],tmp['X2'],tmp['Y'],
                            cmap='rainbow',
                            alpha=0.6
                           )
            ax.set_xlabel('X1')
            ax.set_ylabel('X2')
            ax.set_zlabel('Y')
            ax.view_init(45,-45)
            ax.set_zlim(0.,5.)
            ax.set_title('{0}'.format(string.split('/')[-1].split('_')[0]),fontsize=20)
            
            if save:
                if savefname=='default':
                    string = '/'.join(axeflist[ii].split('/')[0:-1])
                    string += '/{0}_axehelperpoly2dshow3d.png'.format(axeflist[ii].split('/')[-1].split('.')[0])
                else:
                    string = savefname
                fig.savefig(string,bbox_inches='tight')
    ####################
    ####################
    ####################
    def showmask(self,save=False,savefname='default'):
        OBJ = self.obj
        axeflist = self.axehelperbkg.axeflist
        for ii,i in enumerate(OBJ):
            obj = OBJ[i]

            fig = plt.figure(figsize=(10,10))
            ax = plt.axes(projection='3d')
            tmp = copy.deepcopy(obj.data)
            ax.plot_surface(tmp['X1'],tmp['X2'],obj.data['MASKOBJ'].astype(int),
                            cmap='viridis'
                           )
            ax.set_xlabel('X1')
            ax.set_ylabel('X2')
            ax.set_zlabel('MASKOBJ = 1')
            ax.view_init(45,-45)
            ax.set_title('{0}'.format(i.split('/')[-1].split('_')[0]),fontsize=20)

            plt.figure(figsize=(10,10))
            tmp = obj.data['Y']*obj.data['MASKOBJ'].astype(int)
            m = np.where(np.isfinite(tmp))
            vmin,vmax = np.percentile(tmp[m],5.),np.percentile(tmp[m],99.)
            plt.imshow(obj.data['Y']*obj.data['MASKOBJ'].astype(int),
                       cmap='Greys',origin='lower',vmin=vmin,vmax=vmax
                      )

            fig = plt.figure(figsize=(10,10))
            ax = plt.axes(projection='3d')
            tmp = copy.deepcopy(obj.data)
            ax.plot_surface(tmp['X1'],tmp['X2'],obj.model['MASKFIT'].astype(int),
                            cmap='viridis'
                           )
            ax.set_xlabel('X1')
            ax.set_ylabel('X2')
            ax.set_zlabel('MASKFIT = 1')
            ax.view_init(45,-45)
            
            if save:
                if savefname=='default':
                    string = '/'.join(axeflist[ii].split('/')[0:-1])
                    string += '/{0}_axehelperpoly2dshowmask.png'.format(axeflist[ii].split('/')[-1].split('.')[0])
                else:
                    string = savefname
                fig.savefig(string,bbox_inches='tight')
    ####################
    ####################
    ####################
    def show3dfit(self,save=False,savefname='default'):
        OBJ = self.obj
        axeflist = self.axehelperbkg.axeflist
        for ii,i in enumerate(OBJ):
            obj = OBJ[i]
            tmp = obj.data.copy()

            fig = plt.figure()

            ax = fig.add_subplot(1,2,1,projection='3d')
            ax.plot_surface(tmp['X1'],tmp['X2'],obj.model['YFIT'],
                            cmap='Greys',
                            alpha=1.
                           )
            ax.plot_surface(tmp['X1'],tmp['X2'],tmp['Y'],
                            cmap='viridis',
                            alpha=0.6
                           )
            ax.set_xlabel('X1')
            ax.set_ylabel('X2')
            ax.set_zlabel('YFIT')
            ax.view_init(45,-45)
            ax.set_zlim(0.,5.)
            ax.set_title('{0}'.format(i.split('/')[-1].split('_')[0]),fontsize=20)


            ax = fig.add_subplot(1,2,2,projection='3d')
            tmp = obj.data.copy()
            ax.plot_surface(tmp['X1'],tmp['X2'],tmp['Y'] - obj.model['YFIT'],
                            cmap='rainbow'
                           )
            ax.set_xlabel('X1')
            ax.set_ylabel('X2')
            ax.set_zlabel('Y - YFIT')
            ax.view_init(45,-45)
            ax.set_zlim(0.,5.)
            
            if save:
                if savefname=='default':
                    string = '/'.join(axeflist[ii].split('/')[0:-1])
                    string += '/{0}_axehelperpoly2dshow3dfit.png'.format(axeflist[ii].split('/')[-1].split('.')[0])
                else:
                    string = savefname
                fig.savefig(string,bbox_inches='tight')    
    ####################
    ####################
    ####################
    def showsum(self,save=False,savefname='default'):
        # https://pythonmatplotlibtips.blogspot.com/2018/01/add-second-x-axis-at-top-of-figure-python-matplotlib-pyplot.html
        OBJ = self.obj
        HEADERALL = self.axehelperbkg.headerall
        axeflist = self.axehelperbkg.axeflist
        for ii,i in enumerate(OBJ):
            obj = OBJ[i]
            tmpdata = obj.data['Y']
            tmpmask = obj.data['MASK']
            tmpheader = HEADERALL[axeflist[ii]]
            xref,yref = tmpheader['XYREF']
            pixx = tmpheader['PIXX']
            pixy = tmpheader['PIXY']
            ww = tmpheader['WW']
            cc0x,cc0y,cc1x,cc1y = tmpheader['CC']
            sectx = tmpheader['SECTX']
            secty = tmpheader['SECTY']

            plt.figure(figsize=(10,10))

            ax1 = plt.subplot(2,1,1)
            ax1.imshow(obj.model['YSUB'],cmap='Greys',origin='lower',
                       vmin=-1,vmax=1
                      )

            bb0x = sectx[0]+sectx[1]
            bb1x = bb0x+sectx[2]
            bb0y = secty[0]
            bb1y = bb0y+secty[1]+secty[2]+secty[3]
            tmpx = [bb0x,bb1x,bb1x,bb0x,bb0x]
            tmpy = [bb0y,bb0y,bb1y,bb1y,bb0y]
            ax1.plot(tmpx,tmpy,'r-',alpha=0.6)
            ax1.set_frame_on(False)
            ax1.set_title(i.split('/')[-1].split('_')[0],fontsize=20)

            ax2 = plt.subplot(2,1,2,sharex=ax1)
            tmpx = np.arange(0,obj.model['SUM'].shape[0])
            ax2.plot(tmpx,obj.model['SUM'])
            ax2.set_ylabel('SUM',fontsize=20)
            ax2.set_frame_on(False)
            ax2.grid()

            numtick = 10
            ax22 = ax2.twiny()
            ax22.set_frame_on(False)
            tmpp = ww
            tmp = tmpp.astype(int).copy()
            newlabel = np.linspace(tmp[0],tmp[-1],numtick).astype(int)
            newpos = np.linspace(tmpx[0],tmpx[-1],numtick)
            ax22.set_xticks(newpos)
            ax22.set_xticklabels(newlabel)
            ax22.set_xlabel('obs. wavelength (A)',fontsize=15)
            ax22.set_xlim(ax2.get_xlim())

            if save:
                if savefname=='default':
                    string = '/'.join(axeflist[ii].split('/')[0:-1])
                    string += '/{0}_axehelperpoly2dshowsum.png'.format(axeflist[ii].split('/')[-1].split('.')[0])
                else:
                    string = savefname
                plt.savefig(string,bbox_inches='tight')    



# Kornpob Bhirombhakdi

import numpy as np
import matplotlib.pyplot as plt
import copy,glob,os
from astropy.io import fits
from math import pi

class AXEhelper_BKG:
    def __init__(self,axeflist=None,fltflist=None,
                 padxleft=5,padxright=5,
                 padylow=10,halfdy=3,padyup=10,
                 adjusty=1.2
                ):
        self.axeflist = axeflist
        self.fltflist = fltflist
        self.params = (padxleft,padxright,padylow,halfdy,padyup,adjusty)
        self.headerall = self._headerall()
    ####################
    ####################
    ####################
    def make_poly2d(self):
        OBJ = {}
        fltflist = self.fltflist
        axeflist = self.axeflist
        HEADERALL = self.headerall
        for ii,i in enumerate(fltflist):
            # image
            tmp = fits.open(i)
            tmpdata = tmp[1].data.copy()
            tmpdq = tmp['DQ'].data.copy()

            # header & prep
            tmpheader = HEADERALL[axeflist[ii]]
            xref,yref = tmpheader['XYREF']
            pixx = tmpheader['PIXX']
            pixy = tmpheader['PIXY']
            cc0x,cc0y,cc1x,cc1y = tmpheader['CC']
            sectx = tmpheader['SECTX']
            secty = tmpheader['SECTY']

            # Polynomial2D
            x1 = np.arange(cc0x,cc1x)
            x2 = np.arange(cc0y,cc1y)
            x1,x2 = np.meshgrid(x1,x2)

            obj = Polynomial2D()
            obj.data['X1'] = x1.copy()
            obj.data['X2'] = x2.copy()
            obj.data['Y'] = tmpdata[cc0y:cc1y,cc0x:cc1x]
        #     print(x1.shape,x2.shape,obj.data['Y'].shape)    

            # data['MASK']
            tmp = np.full_like(tmpdq,True,dtype=bool)
            m = np.where(tmpdq==0)
            tmp[m] = False
            obj.data['MASK'] = tmp[cc0y:cc1y,cc0x:cc1x]
        #     print(obj.data['Y'].shape,obj.data['MASK'].shape)  

            OBJ[i] = copy.deepcopy(obj)
        return OBJ
    ####################
    ####################
    ####################
    def _headerall(self):
        axeflist = self.axeflist
        fltflist = self.fltflist
        padxleft,padxright,padylow,halfdy,padyup,adjusty = self.params
        
        tmp = {}
        for i in axeflist:
            # read from header
            HEADER = copy.deepcopy(fits.open(i)[1].header)
            xref,yref = HEADER['REFPNTX'],HEADER['REFPNTY']
            bb0x,bb1x = HEADER['BB0X'],HEADER['BB1X']
            orient = HEADER['ORIENT']
            cpointx,cpointy = HEADER['CPOINTX'],HEADER['CPOINTY']
            dldx0,dldx1 = HEADER['DLDX0'],HEADER['DLDX1']

            # manually adjust offset
            yref += adjusty

            # trace and wavelength
            fny = lambda x : np.tan((90.+orient)*pi/180.) * (x - cpointx) + yref
            fnw = lambda x : dldx1 * (x - cpointx) + dldx0

            pixx = np.array([round(xref),round(bb1x)],dtype=int)
            pixy = np.round(fny(pixx)).astype(int)
            ww = fnw(pixx)
            
            # section
            pixywidth = pixy[-1] - pixy[0] + 1
            sectx = (padxleft,round(bb0x-xref),round(bb1x-bb0x),padxright)
            secty = (padylow,halfdy,pixywidth,halfdy,padyup)

            # cut box
            cc0x = round(xref)-padxleft
            cc1x = round(bb1x)+padxright
            cc0y = int(fny(cc0x))-halfdy-padylow
            cc1y = int(fny(cc1x))+halfdy+padyup

            # output 
            tmp[i] = {}
            tmp[i]['XYREF'] = (xref,yref)
            tmp[i]['DLDX'] = (dldx0,dldx1)
            tmp[i]['BBX'] = (bb0x,bb1x)
            tmp[i]['PIXX'] = pixx.copy()
            tmp[i]['PIXY'] = pixy.copy()
            tmp[i]['WW'] = ww.copy()
            tmp[i]['SECTX'] = copy.deepcopy(sectx)
            tmp[i]['SECTY'] = copy.deepcopy(secty)
            tmp[i]['CC'] = (cc0x,cc0y,cc1x,cc1y)

        return copy.deepcopy(tmp)
    ####################
    ####################
    ####################
    def show(self,save=False,savefname='default'):
        fltflist = self.fltflist
        axeflist = self.axeflist
        HEADERALL = self.headerall

        for ii,i in enumerate(fltflist):
            tmp = fits.open(i)
            tmpdata = tmp[1].data.copy()

            tmpheader = HEADERALL[axeflist[ii]]
            xref,yref = tmpheader['XYREF']
            pixx = tmpheader['PIXX']
            pixy = tmpheader['PIXY']
            ww = tmpheader['WW']
            cc0x,cc0y,cc1x,cc1y = tmpheader['CC']
            sectx = tmpheader['SECTX']
            secty = tmpheader['SECTY']

            fig,ax = plt.subplots(2,1,sharex=True)
            fig.tight_layout()
            m = np.where(np.isfinite(tmpdata))
            vmin,vmax = np.percentile(tmpdata[m],5.),np.percentile(tmpdata[m],99.)
            ax[0].imshow(tmpdata,origin='lower',cmap='viridis',vmin=vmin,vmax=vmax)
            ax[0].scatter(xref,yref,s=30,facecolor='red',edgecolor='None')
            ax[0].plot(pixx,pixy,'r-')
            ax[0].set_xlim(cc0x,cc1x)
            ax[0].set_ylim(cc0y,cc1y)
            ax[0].set_title('{0}'.format(i.split('/')[-1].split('_')[0]),fontsize=20)
            ax[0].set_ylabel('pixY',fontsize=20)

            bb0x = cc0x+sectx[0]+sectx[1]
            bb1x = bb0x+sectx[2]
            bb0y = cc0y+secty[0]
            bb1y = bb0y+secty[1]+secty[2]+secty[3]
            tmpx = [bb0x,bb1x,bb1x,bb0x,bb0x]
            tmpy = [bb0y,bb0y,bb1y,bb1y,bb0y]
            ax[0].plot(tmpx,tmpy,'r-')

            ax[1].plot(pixx,ww)
            ax[1].set_xlabel('pixX',fontsize=20)
            ax[1].set_ylabel('obs. wavelength (A)',fontsize=20)
            ax[1].grid()
            
            if save:
                if savefname=='default':
                    string = '/'.join(axeflist[ii].split('/')[0:-1])
                    string += '/{0}_axehelperbkg.png'.format(axeflist[ii].split('/')[-1].split('.')[0])
                else:
                    string = savefname
                fig.savefig(string,bbox_inches='tight')
            
    


import numpy as np

def read_postaXecombine(combinefile):
    """
    This function reads combine.dat from post-aXe processing. combine.dat is a text file with the following structure: first line = '# LAMBDA,FLAM,FLAM_ORIG,FERROR' and the following lines are csv entries.
    """
    filepath = combinefile
    f = open(filepath,'r')
    keys = []
    tmp = np.array([])
    for i,ii in enumerate(f.readlines()):
        if i==0:
            keys = np.array(ii.split(' ')[1].split(','))
        else:
            tmp = np.concatenate((tmp,np.array(ii.split(','))))
    nrow = int(len(tmp)/len(keys))
    ncol = len(keys)
    tmp = np.reshape(tmp,(nrow,ncol))
    out = {}
    for i,ii in enumerate(keys):
        out[ii] = tmp[:,i].astype(float)
    return out



# Kornpob Bhirombhakdi 20200331

from astropy.io import fits
import numpy as np

def order_flist(flist,key=None):
    """
    Input:
      flist = flt.fits file list
      key = keyword from PRIMARY hdu for ordering
    Output: a list of filenames ordered by using key
    ##########
    If not key, key = 'EXPSTART' by default
    """
    if not key:
        key = 'EXPSTART'
    tmp = []
    for i,ii in enumerate(flist):
        x = fits.open(ii)
        tmpp = x[0].header[key]
        tmp.append(tmpp)
    order = np.argsort(tmp)
    return np.array(flist)[order]



# Kornpob Bhirombhakdi 20200331

import numpy as np
import pandas as pd

def axeapcorr(aXeSPCdata,apcorrmod,outputpath=None):
    """
    Input:
      aXeSPCdata = data from SPC or opt.SPC output files from aXe reduction. 
      apcorrmod = model for aperture correction (i.e., apcorr = f(wavelength, apsize)). 
      outputpath = filepath to save output from axeapcorr.
    Output: csv file saved to outputpath. Output is aperture-size corrected.
    ##########
    axeapcorr performs aperture correction of the SPC or opt.SPC outputs from aXe reduction, and saves the results to a specified path. Aperture corrections for aXe are tables presented in ISRs. These tables are wrapped in grismapcorr.py.
    """
    # set up variables
    x = aXeSPCdata
    lamb = x['LAMBDA']
    flux = x['FLUX']
    eflux = x['FERROR']
    tcount = x['TCOUNT']
    weight = x['WEIGHT']
    dlambda = x['DLAMBDA']
    apsize = (weight/2.) * 0.13 # 0.13 is pixel scale of WFC3/IR
    
    # calculate aperture correction
    APCORR = []
    for i,ii in enumerate(lamb):
        tmp = apcorrmod(lamb[i],apsize[i])
        APCORR.append(tmp[0])
    APCORR = np.array(APCORR)
    
    # perform correction
    tmpp,etmpp = flux/APCORR,eflux/APCORR
    tmppp = tcount/APCORR/dlambda
    
    # prepare output
    out =  {'APCORR': APCORR,
            'FLUX': tmpp,
            'FLUX_ORIG': flux,
            'FERROR': etmpp,
            'TCOUNT': tmppp,
            'TCOUNT_ORIG': tcount,
            'TERROR': None,
            'U_TCALIB': 'count_unit / A',
            'U_TORIG': 'count_unit / pix',
            'DLAMBDA': dlambda,
            'U_DLAMBDA': 'A / pix',
            'LAMBDA': lamb}
    
    # save output
    _axeapcorrsave(out,outputpath)
    return out

def _axeapcorrsave(out,outputpath):
    if not outputpath:
        outputpath = 'axeapcorrsave.csv'
    pd.DataFrame(out).to_csv(outputpath)



# Kornpob Bhirombhakdi 20200331
# This is a very specific helper for aXe to go with a new version of sextractor.

import pandas as pd
import copy
import numpy as np

def read_catalog(catalog):
    x = open(catalog,'r')
    colname = []
    values = []
    for i in x.readlines():
        if i[0]=='#':
            colname.append(i)
        else:
            x2 = np.array(i.split(),dtype=np.double)
            values.append(x2)
    x.close()
    x3 = pd.DataFrame(values,columns=colname)
    x3 = x3.astype({colname[2]:'int64'})
    return copy.deepcopy(x3)

def change_magiso2magwavelength(catalog,wavelength):
    x = read_catalog(catalog)
    y = x.columns[11].split()
    y[2] = 'MAG_F'+str(int(wavelength))
    z = ' '.join(y) + '\n'
    x.rename(columns={x.columns[11]:z},inplace=True)
    write_catalog(catalog,x)
    
def write_catalog(catalog,dataframe):
    x = open(catalog,'w')
    for i in dataframe.columns:
        x.writelines(i)
    for i in dataframe.values:
        string = ' '.join(i[0:2].astype(str))
        string = string + ' ' + str(i[2].astype(int))
        string = string + ' ' + ' '.join(i[3:].astype(str)) + '\n'
        x.writelines(string)
    x.close()

def change_catalog_order(catalog):
    x = read_catalog(catalog)
    y = x.T.iloc[0:14].copy()
    z = y.copy()
    z1,z11 = z.iloc[11].name,z.iloc[12].name
    zz,zz2 = z1.split(),z11.split()
    zz[1],zz2[1] = '13','12'
    z2 = ' '.join(zz) + '\n'
    z22 = ' '.join(zz2) + '\n'
    z.rename(index={z1:z2,z11:z22}, inplace = True)
    tmp = z.iloc[0:11]
    tmp = tmp.append(z.iloc[12])
    tmp = tmp.append(z.iloc[11])
    tmp = tmp.append(z.iloc[13])
    tmpp = tmp.T.copy()
    write_catalog(catalog,tmpp)
    
def remove_bad_entry(catalog,maglimit=27.):
    x = read_catalog(catalog)
    for i in x.columns:
        for j in i.split():
            if j.split('_')[0] == 'MAG':
                xx = i
                break
    tmp = x[x[xx]<maglimit].copy()
    write_catalog(catalog,tmp)

def one_source(catalog,ID):
    x = read_catalog(catalog)
    for i in x.columns:
        for j in i.split():
            if j == 'NUMBER':
                xx = i
                break
    df = x[x[xx]==ID].copy()
    return df
    
def select_source(v1,v2,catalog,method='RADEC'):
    # check variables
    if method in {'RADEC','XY'}:
        if not v2:
            raise ValueError('v2 must be specified given method.')
    if method=='RADEC':
        CATALOG = read_catalog(catalog)
        x0 = CATALOG.iloc[:,3].values
        y0 = CATALOG.iloc[:,4].values
        dx,dy = x0-v1,y0-v2
        dd = np.sqrt(dx**2 + dy**2)
        index = np.argmin(dd)
    elif method=='XY':
        CATALOG = read_catalog(catalog)
        x0 = CATALOG.iloc[:,0].values
        y0 = CATALOG.iloc[:,1].values
        dx,dy = x0-v1,y0-v2
        dd = np.sqrt(dx**2 + dy**2)
        index = np.argmin(dd)
    elif method=='ID':
        CATALOG = read_catalog(catalog)
        x0 = CATALOG.iloc[:,2].values
        dx = x0-v1
        dd = np.sqrt(dx**2)
        index = np.argmin(dd)
    else:
        raise ValueError('select_source(method) is not allowed.')
    return CATALOG.iloc[index,:]



# Kornpob Bhirombhakdi
# kbhirombhakdi@stsci.edu

from astropy.io import fits
import copy
import numpy as np

class AXEhelper_FlatField:
    '''
    AXEhelper_FlatField (using Python 3) computes a flatfield image in aXe definition.
    This can easily be implemented following the use of AXEhelper_computeTraceNWavelength.
    Given obj = AXEhelper_computeTraceNWavelength(...), and obj.compute() is called,
        a user can start with flat = AXEhelper_FlatField(flatfile,obj.wavelength['WW'])
    flat.flatfile to see the path to flatfile.
    flat.data to see inputs for the computation.
    flat.compute() to compute the flatfield image.
    flat.flatfield to see the flatfield image.
    '''
    def __init__(self,flatfile,ww):
        self.flatfile = flatfile
        self.data = {'WMIN':None,
                     'WMAX':None,
                     'FLAT':None,
                     'NAX1':None,
                     'NAX2':None,
                     'WW':ww
                    }
        self.flatfield = None
        try:
            self._get()
        except:
            pass
    def _get(self):
        tmp = fits.open(self.flatfile)
        wmin,wmax = tmp[0].header['WMIN'],tmp[0].header['WMAX']
        nax1,nax2 = tmp[0].header['NAXIS1'],tmp[0].header['NAXIS2']
        self.data['WMIN'] = wmin
        self.data['WMAX'] = wmax
        self.data['NAX1'] = nax1
        self.data['NAX2'] = nax2
        n = len(tmp)
        tmpp = {}
        for i in range(n):
            tmpp[i] = tmp[i].data.copy()
        self.data['FLAT'] = copy.deepcopy(tmpp)
    ##########
    ##########
    ##########
    def compute(self):
        # dimension test
        if self.data['WW'].shape[0] != self.data['NAX1']:
            raise ValueError('Dimension of WW != NAX1')
        # set parametric wavelength [0,1]
        wmin,wmax,ww = self.data['WMIN'],self.data['WMAX'],self.data['WW']
        paramww = (ww-wmin)/(wmax-wmin)
        paramww[paramww<0.] = 0.
        paramww[paramww>1.] = 1.
        paramww,tmp = np.meshgrid(paramww,np.arange(0,self.data['NAX2']))
        # compute
        tmp = self.data['FLAT']
        tmpp = np.full_like(paramww,0.,dtype=float)
        for i in tmp:
            tmpp += tmp[i] * np.power(paramww,i)
        self.flatfield = tmpp.copy()
    